import path from "node:path";
import { parsePlaybillComponent } from "define";
import YAML, { YAMLParseError } from "yaml";
import { ZodError } from "zod";
import { loadFileOrFail } from "#util";
import { MalformedResourceFileError } from "./errors";
import { toSlug } from "./slug";
import type { AnyPlaybillComponent } from "@proscenium/playbill";

type FileFormat = "yaml" | "markdown";

const FRONTMATTER_REGEX = /^---[\s\S]*?---/gm;

/**
 * Attempt to parse YAML frontmatter from a mixed yaml/md doc
 * @param content The raw markdown content
 * @returns Any successfully parsed frontmatter
 */
function extractFrontmatter(content: string) {
  // If it does not start with `---`, it is invalid for frontmatter
  if (content.trim().indexOf("---") !== 0) {
    return {};
  }

  if (FRONTMATTER_REGEX.test(content)) {
    const frontmatterString = content.match(FRONTMATTER_REGEX)?.[0] ?? "";
    const cleanFrontmatter = frontmatterString.replaceAll("---", "").trim();
    return YAML.parse(cleanFrontmatter);
  }

  return {};
}

/**
 * Given a string of a markdown document, extract the markdown content
 * @param content The raw markdown content
 * @returns The markdown content without frontmatter
 */
function extractMarkdown(content: string): string {
  if (content.trim().indexOf("---") !== 0) {
    return content;
  }
  return content.replace(FRONTMATTER_REGEX, "").trim();
}

function parseYAMLResourceFile(
  filePath: string,
  text: string,
): AnyPlaybillComponent[] {
  const parsedDocs = YAML.parseAllDocuments(text);

  if (parsedDocs.some((doc) => doc.toJS() === null)) {
    throw new MalformedResourceFileError("Encountered NULL resource", filePath);
  }

  const errors = parsedDocs.flatMap((doc) => doc.errors);

  if (errors.length > 0) {
    throw new MalformedResourceFileError(
      "Error parsing YAML resource",
      filePath,
      errors.map((e) => e.message).join(", "),
    );
  }

  const collection: AnyPlaybillComponent[] = [];

  for (const doc of parsedDocs) {
    const raw = doc.toJS();
    const parsedComponent = parsePlaybillComponent(raw);
    if (parsedComponent !== null) {
      collection.push(parsedComponent);
    }
  }

  return collection;
}

function parseMarkdownResourceFile(
  filePath: string,
  text: string,
): AnyPlaybillComponent[] {
  try {
    const defaultName = path.basename(filePath, ".md");
    const defaultId = toSlug(defaultName);
    const frontmatter = extractFrontmatter(text);
    const markdown = extractMarkdown(text);

    const together = parsePlaybillComponent({
      // Use the file name, allow it to be overridden by frontmatter
      name: defaultName,
      id: defaultId,
      ...frontmatter,
      description: markdown,
    });

    // Null means hidden
    if (together === null) {
      return [];
    }

    return [together];
  } catch (e) {
    if (e instanceof YAMLParseError) {
      throw new MalformedResourceFileError(
        "Error parsing Markdown frontmatter",
        filePath,
        e.message,
      );
    }

    if (e instanceof ZodError) {
      throw new MalformedResourceFileError(
        "Error parsing resource file",
        filePath,
        e.message,
      );
    }

    throw e;
  }
}

export class ComponentFile {
  private _filePath: string;
  private _raw = "";
  private _format: FileFormat;
  private _components: AnyPlaybillComponent[] = [];
  private _basename: string;

  constructor(filePath: string) {
    this._filePath = path.resolve(filePath);

    const extension = path.extname(filePath).slice(1).toLowerCase();
    this._basename = path.basename(filePath, `.${extension}`);

    switch (extension) {
      case "yaml":
        this._format = "yaml";
        break;
      case "md":
        this._format = "markdown";
        break;
      default:
        throw new Error(`Unsupported file format: ${extension}`);
    }
  }

  async load() {
    this._raw = await loadFileOrFail(this._filePath);
    switch (this._format) {
      case "yaml":
        this._components = parseYAMLResourceFile(this._filePath, this._raw);
        break;
      case "markdown":
        this._components = parseMarkdownResourceFile(this._filePath, this._raw);
    }
  }

  /**
   * An array of well-formed components defined in this file
   */
  get components() {
    return this._components;
  }

  /**
   * The aboslute file path of this file
   */
  get path() {
    return this._filePath;
  }

  get baseName() {
    return this._basename;
  }
}
