import rehypeStringify from "rehype-stringify";
import remarkParse from "remark-parse";
import remarkRehype from "remark-rehype";
import { unified } from "unified";
import type { BoundPlaybill } from "./binding";
import remarkWikiLink from "remark-wiki-link";
import { toSlug } from "./slug";
import rehypeRaw from "rehype-raw";
import remarkGfm from "remark-gfm";
import rehypeShiftHeading from "rehype-shift-heading";
import type { AnyPlaybillComponent } from "@proscenium/playbill";

export type MarkdownParserFunction = (input: string) => Promise<string>;

export function createMarkdownRenderer(
  binding: BoundPlaybill,
): MarkdownParserFunction {
  // TODO: Allow specifying links to other files via file paths
  // In this scenario, assume its a markdown-defined file and find the ID, use that for the link
  // to allow for more natural Obsidian-like linking
  // For now, this will mostly work, but if you need to specify a unique ID, it wont work in Obsidian
  // despite being valid for Muse

  const parser = unified()
    .use(remarkParse)
    .use(remarkGfm)
    .use(remarkWikiLink, {
      aliasDivider: "|",
      permalinks: binding.playbill.allComponents().map((c) => c.id),
      pageResolver: (permalink: string) => {
        return [toSlug(permalink)];
      },
      hrefTemplate: (permalink: string) => `#component/${permalink}`,
    })
    .use(remarkRehype, { allowDangerousHtml: true })
    .use(rehypeRaw)
    .use(rehypeShiftHeading, { shift: 1 })
    .use(rehypeStringify);

  // TODO: Add sanitization

  return async (input: string): Promise<string> => {
    const parsed = await parser.process(input);
    return String(parsed);
  };
}

export async function compileMarkdownInPlaybill(
  boundPlaybill: BoundPlaybill,
): Promise<BoundPlaybill> {
  const renderMarkdown = createMarkdownRenderer(boundPlaybill);
  const playbill = boundPlaybill.playbill;

  // Define a processor function to iterate over all components and process their markdown
  const processMarkdownInComponent = async (entry: AnyPlaybillComponent) => {
    entry.description = await renderMarkdown(entry.description);

    if (entry._component === "resource" && entry.type === "table") {
      const newData: string[][] = [];
      for (const row of entry.data) {
        const newRow: string[] = [];
        for (const cell of row) {
          newRow.push(await renderMarkdown(cell));
        }
        newData.push(newRow);
      }
      entry.data = newData;
    }
  };

  // Process all components
  await playbill.processComponents(processMarkdownInComponent);

  return boundPlaybill;
}
