import path from "node:path";
import YAML from "yaml";
import { FileError, FileNotFoundError } from "../lib/errors";

/**
 * Load a file from the filesystem or throw an error if it does not exist.
 * @param filePath The path to the file to load
 * @returns The contents of the file as a string
 *
 * @throws {FileNotFoundError} if the file does not exist
 */
export async function loadFileOrFail(filePath: string): Promise<string> {
  const fileToLoad = Bun.file(filePath);
  const fileExists = await fileToLoad.exists();

  if (!fileExists) {
    throw new FileNotFoundError(`File not found: ${filePath}`);
  }

  return await fileToLoad.text();
}

/**
 * Load and parse a YAML file or throw an error if the file doesnt exist
 * or the yaml fails to parse
 * @param filePath The path to the file to load
 * @returns The parsed contents of the file
 *
 * @throws {FileNotFoundError} if the file does not exist
 * @throws {FileError} if the file is not valid YAML
 */
export async function loadYAMLFileOrFail(filePath: string): Promise<unknown> {
  try {
    return YAML.parse(await loadFileOrFail(filePath));
  } catch (error) {
    throw new FileError("Failed to parse YAML file", filePath, `${error}`);
  }
}

/**
 * Load and parse a JSON file or throw an error if the file doesnt exist
 * or the JSON fails to parse
 * @param filePath The path to the file to load
 * @returns The parsed contents of the file
 *
 * @throws {FileNotFoundError} if the file does not exist
 * @throws {FileError} if the file is not valid JSON
 */
export async function loadJSONFileOrFail(filePath: string): Promise<unknown> {
  try {
    return JSON.parse(await loadFileOrFail(filePath));
  } catch (error) {
    throw new FileError("Failed to parse JSON file", filePath, `${error}`);
  }
}

/**
 * Returns the absolute path of the directory containing the given file
 *
 * @param filePath The path to the file
 * @returns The absolute path of the directory containing the file
 */
export function getAbsoluteDirname(filePath: string): string {
  return path.resolve(path.dirname(filePath));
}
